/*
 * ImageWindow.cs: Main window of Pastemage
 * Copyright (c) 2004 Daniel Serpell
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using Gtk;
using Vector3D;

public class ImageWindow {
    
    // Gtk window
    private Gtk.Window win;
    
    // Glade widgets
    [Glade.Widget]
    private Gtk.DrawingArea daImageView;
    
    [Glade.Widget]
    private Gtk.SpinButton sbImageAngle;
    
    [Glade.Widget]
    private Gtk.Button btnImageInfo;
    
    // Image Info
    public class imageInfo {
        public Gdk.Pixbuf pix;
        public v3d N, U, V;
        public double scale;
        public int sx,sy;
    };
    
    private ViewData view;
    
    private System.Collections.ArrayList images;
    private int selectedImage;
    private double pressX, pressY;
    
    private TileCache tileCache;
    
    public ImageWindow ()
    {
    
        Glade.XML xml = new Glade.XML ("pastemage.glade", "ImageWindow", null);
        xml.Autoconnect(this);
        win = (Gtk.Window)xml.GetWidget("ImageWindow");
        
        // Create image list
        images = new System.Collections.ArrayList();
        selectedImage = -1;
        
        // Init view
        view = new ViewData();
        view.goNorth();
        view.sx = daImageView.Allocation.Width;
        view.sy = daImageView.Allocation.Height;
        view.NotifyUpdate += OnViewUpdated;
        
        // Init tile-cache
        tileCache = new TileCache();
        
        // Activate some events
        daImageView.AddEvents( (int)Gdk.EventMask.ButtonMotionMask );
        daImageView.AddEvents( (int)Gdk.EventMask.ButtonPressMask );
        daImageView.AddEvents( (int)Gdk.EventMask.ScrollMask );
        daImageView.AddEvents( (int)Gdk.EventMask.StructureMask );
    }
    
    private unsafe int getImgData(Gdk.Pixbuf p, int x, int y)
    {
        if( x < 0 || y < 0 || x > p.Width || y > p.Height )
            return 0;
        byte *c = (byte *)p.Pixels;
        c += (p.Rowstride * y + x * 3);
        return c[0] + c[1] * 256 + c[2] * 65536;
    }
    
    // Handlers...
    void OnViewUpdated(object o, EventArgs args)
    {
        tileCache.invalidateView();
        daImageView.QueueDraw();
    }
    
    void OnPreferences(object o, EventArgs args)
    {
        new Preferences(view);
    }
    
    void OnQuit (object o, DeleteEventArgs args)
    {
        Application.Quit ();
    }
    
    void OnQuit (object o, EventArgs args)
    {
        Application.Quit ();
    }
    
    void OnAbout (object o, EventArgs args)
    {
        // TODO: A new about window...
        MessageDialog m = new MessageDialog(win,
            DialogFlags.DestroyWithParent, MessageType.Info,
            ButtonsType.Close,
            "Pastemage: A program to join panorama images.\n"+
            "(c) 2004 Daniel Serpell");
    }
    
    void OnHome(object o, EventArgs args)
    {
        // Init view
        view.goNorth();
        view.Zoom = 1;
        tileCache.invalidateView();
        daImageView.QueueDraw();
    }
    
    void OnZoomIn(object o, EventArgs args)
    {
        view.Zoom *= 1.25;
        tileCache.invalidateView();
        daImageView.QueueDraw();
    }
    
    void OnZoomOut(object o, EventArgs args)
    {
        view.Zoom *= 0.8;
        tileCache.invalidateView();
        daImageView.QueueDraw();
    }
    
    void OnSavePanorama(object o, EventArgs args)
    {
        // TODO.
    }
    
    void OnAdd(object o, EventArgs args)
    {
        CompatFileChooserDialog fs = new CompatFileChooserDialog("Select a file to add", win, CompatFileChooserDialog.Action.Open);
        //Gtk.FileSelection fs = new FileSelection("Select a file to add");
        
        if( fs.Run() != (int)(Gtk.ResponseType.Ok) )
        {
            fs.Destroy();
            return;
        }
        
        string fileName = fs.Filename;
        fs.Destroy();
        // Try to load the image
        try
        {
            Gdk.Pixbuf p = new Gdk.Pixbuf(fileName);
            imageInfo i = new imageInfo();
            i.pix = p;
            i.N.x = 0;
            i.N.y = 0;
            i.N.z = 1;
            i.U.x = 1;
            i.U.y = 0;
            i.U.z = 0; 
            i.V.x = 0;
            i.V.y = 1;
            i.V.z = 0;
            i.scale = 5.4 / 0.00258;
            i.sx = p.Width;
            i.sy = p.Height;  
            images.Add( i );
            tileCache.invalidateImage();
            unselectImage();
        }
        catch( Exception e )
        {
            MessageDialog m = new MessageDialog(win,
                DialogFlags.DestroyWithParent, MessageType.Error,
                ButtonsType.Close, "Error loading file: " + e.ToString());
            m.Run();
            m.Destroy();
        }
    }
    
    void OnDaImageViewScroll(object o, ScrollEventArgs args)
    {
	if( args.Event.Direction == Gdk.ScrollDirection.Up )
	{
	    view.Zoom *= 1.25;
            tileCache.invalidateView();
	    daImageView.QueueDraw();
	}
	else if( args.Event.Direction == Gdk.ScrollDirection.Down )
	{
	    view.Zoom *= 0.8;
            tileCache.invalidateView();
	    daImageView.QueueDraw();
	}
    }
    
        
    void OnDaImageViewConfigure(object o, ConfigureEventArgs args)
    {
        // Configure!
        view.sx = args.Event.Width;
        view.sy = args.Event.Height;
        tileCache.invalidateView();
    }
        
    unsafe void OnDaImageViewExpose(object o, ExposeEventArgs args)
    {
        if( ! tileCache.ValidView() )
            tileCache.updateViewCache(view);
        if( ! tileCache.ValidImage() )
            tileCache.updateImageCache(images);
        
        // The tile size
        int ts = view.TileSize;
        
        // Create a buffer to fill with tile image data.
        byte[] pb = new byte[ts*ts*3];
        // An the GC
        Gdk.GC gc = new Gdk.GC(daImageView.GdkWindow);
        
        // Now, paint each tile in turn.
        int nx = tileCache.getTilesX();
        int ny = tileCache.getTilesY();
        
        for(int ty=0;ty<ny;ty++)
        {
            for(int tx=0;tx<nx;tx++)
            {
                // Paint the tile (tx,ty)
                int n = images.Count;
                // Buffers to fill tile image data
                int[,] red   = new int[ts,ts];
                int[,] green = new int[ts,ts];
                int[,] blue  = new int[ts,ts];
                int[,] count = new int[ts,ts];
                
                // For each image, iterate all pixels in the tile
                for(int i=0;i<n;i++)
                {
                    TileCache.ImageInter ii = tileCache.tiles[tx,ty].imgInterArray[i];
                    imageInfo im = (imageInfo)images[i];
                    if( ! ii.visible )
                        continue;
                    
                    double u0 = ii.ax;
                    double v0 = ii.ay;
                    double u1 = ii.bx;
                    double v1 = ii.by;
                    for(int y=0;y<ts;y++)
                    {
                        double u = u0;
                        double v = v0;
                        for(int x=0;x<ts;x++,u+=u1,v+=v1)
                        {
                            if( (u < 0) || (u >= im.sx) )
                                continue; 
                            if( (v < 0) || (v >= im.sy) )
                                continue;
                            byte *val = (byte *)im.pix.Pixels;
                            val += (im.pix.Rowstride * ((int)v) + ((int)u) * 3);
                            if( i == selectedImage )
                            {
                                count[x,y] += 2;
                                red[x,y]   += (val[0]*3 + 255)/2;
                                green[x,y] += (val[1]*3 + 255)/2;
                                blue[x,y]  += (val[2]*3 + 255)/2;
                            }
                            else
                            {
                                count[x,y] += 1;
                                red[x,y]   += val[0];
                                green[x,y] += val[1];
                                blue[x,y]  += val[2];
                            }
                        }
                        u0 += ii.cx;
                        v0 += ii.cy;
                        u1 += ii.dx;
                        v1 += ii.dy;
                    }
                }
                // Now, copy the tile to the byte buffer
                int bp = 0;
                for(int y=0;y<ts;y++)
                {
                    for(int x=0;x<ts;x++)
                    {
                        if( count[x,y] == 0 )
                        {
                            byte c = (byte)(63 + (((tx*ts)^(ty*ts))&64));
                            pb[bp] = c;
                            pb[bp+1] = c;
                            pb[bp+2] = c;
                        }
                        else
                        {
                            pb[bp+0] = (byte)(red[x,y]/count[x,y]);
                            pb[bp+1] = (byte)(green[x,y]/count[x,y]);
                            pb[bp+2] = (byte)(blue[x,y]/count[x,y]);
                        }
                        bp += 3;
                    }
                }
                // Copy the byte buffer to the screen
                daImageView.GdkWindow.DrawRgbImage(gc,tx*ts,ty*ts,ts,ts,Gdk.RgbDither.Normal,pb,3*ts);
            }    
        }
        
    }
    
    void OnDaImageViewButtonPress(object o, ButtonPressEventArgs args)
    {
	// Store new button coordinates
        double x = args.Event.X;
        double y = args.Event.Y;
        pressX = x;
        pressY = y;

	// Select an image if button 1 is pressed
	if( (args.Event.Button == 1) && 
	    (0 == (args.Event.State & Gdk.ModifierType.ShiftMask)) )
	{
            // See if there is an image under the cursor
            int i;
            int n = images.Count;
            // Get's the size of the window
            int sx = daImageView.Allocation.Width;
            int sy = daImageView.Allocation.Height;
            // And the min size, to cal. the view angle.
	    double zoom;
	    if( sx < sy )
		zoom = view.Zoom / sx;
	    else
		zoom = view.Zoom / sy;
        
            for(i=0;i<n;i++)
	    {
	        imageInfo im = (imageInfo)images[i];
		v3d vw = view.Dir +
		    (x-sx/2)*zoom*view.Dx +
		    (y-sy/2)*zoom*view.Dy;
                double a = v3d.dot(im.N,vw);
	        if( a <= 0 )
	            continue;
		a = im.scale / a;
                double u = a*v3d.dot(im.U,vw);
	        if( (u > im.sx/2) || (u < -im.sx/2) )
	            continue; 
		double v = a*v3d.dot(im.V,vw);
                if( (v > im.sy/2) || (v < -im.sy/2) )
	            continue; 
	        // Ok, this is the image!
		break;
            }
	    if( i < n )
	        selectImage(i);
            else
	        unselectImage();
        }
    }

    void selectImage(int num)
    {
        selectedImage = num;
        imageInfo img = (imageInfo)images[num];
        btnImageInfo.Label = "Selected: " + num.ToString() + " - " +
            img.sx.ToString() + "x" + img.sy.ToString();
        sbImageAngle.Value = 35.0 * img.scale / img.sx ;
        sbImageAngle.Sensitive = true;
        daImageView.QueueDraw();
    }

    void unselectImage()
    {
        selectedImage = -1;
        btnImageInfo.Label = "No image selected";
        sbImageAngle.Value = 0;
        sbImageAngle.Sensitive = false;
        daImageView.QueueDraw();
    }
    
    void OnSbImageAngleValueChanged(object o, EventArgs args)
    {
        if( selectedImage >= 0 )
        {
            imageInfo img = (imageInfo)images[selectedImage];
            double sc = sbImageAngle.Value * img.sx / 35.0;
            if( img.scale != sc )
            {
                img.scale = sc;
                tileCache.invalidateImage();
                daImageView.QueueDraw();
            }
        }
    }

    void OnDaImageViewMotionNotify(object o, MotionNotifyEventArgs args)
    {
        // Get's the size of the window
	int sx = daImageView.Allocation.Width;
	int sy = daImageView.Allocation.Height;
	// And the min size, to cal. the view angle.
	double zoom;
	if( sx < sy )
	    zoom = view.Zoom / sx;
	else
	    zoom = view.Zoom / sy;
	
        // Projection of current mouse coordinates, using view parameters
        double x0 = pressX,       y0 = pressY,
               x1 = args.Event.X, y1 = args.Event.Y;
	
	// Store current coordinates
        pressX = args.Event.X;
        pressY = args.Event.Y;
	
	// If Shift is NOT pressed, move selected image
        if( 0 == (args.Event.State & Gdk.ModifierType.ShiftMask) )
	{
	    // "FROM" and "TO" vectors in the view-plane
            v3d v1 = tileCache.viewVector(x1,y1);
            v3d v0 = tileCache.viewVector(x0,y0);
	    
            if( selectedImage < 0 )
	        return;
            
	    imageInfo im = (imageInfo)images[selectedImage];
	    v3d n0,n1,axis;

	    if( 0 != (args.Event.State & Gdk.ModifierType.Button1Mask) )
            {
	        // Button 1: Rotation axis is normal to the moving vector.
                axis = v3d.cross(v0,v1);
	        axis.normalize();
	    }
            else
	    {
	        // Button 2: Rotation axis is the image normal.
                axis = im.N;
	        axis.normalize();
	        double a = v3d.dot(im.N,v0);
		if( a <= 0 )
                    return;
	        v0 = (1.0/a) * v0 - im.N;
	        a = v3d.dot(im.N,v1);
		if( a <= 0 )
                    return;
	        v1 = (1.0/a) * v1 - im.N;
	    }  
            // Normalized, ie, directions in the view sphere.
            v0.normalize();
	    v1.normalize();
	    // Third axis, normal to the other two
            n0 = v3d.cross(v0,axis);
	    n1 = v3d.cross(v1,axis);
	    // Now, do the rotation of the image vectors.
            im.N = v3d.dot(im.N,v0)*v1+v3d.dot(im.N,n0)*n1+v3d.dot(im.N,axis)*axis;
	    im.U = v3d.dot(im.U,v0)*v1+v3d.dot(im.U,n0)*n1+v3d.dot(im.U,axis)*axis;
	    im.V = v3d.dot(im.V,v0)*v1+v3d.dot(im.V,n0)*n1+v3d.dot(im.V,axis)*axis;
	    // Update the view
            tileCache.invalidateImage();
            daImageView.QueueDraw();
        
	}
	else
	{
	    // Move the view, from v0 to v1
	    v3d v0,v1,axis,n1,n0;
	    
	    if( 0 != (args.Event.State & Gdk.ModifierType.Button1Mask) )
            {
	        // Button 1: Rotation axis is normal to the moving vector.
                v0 = tileCache.viewVector(x1,y1);
                v1 = tileCache.viewVector(x0,y0);
                axis = v3d.cross(v0,v1);
	        axis.normalize();
	    }
            else
	    {
	        // Button 2: Rotation axis is the view center.
                v0 = tileCache.viewDirection(x1,y1);
                v1 = tileCache.viewDirection(x0,y0);
                axis = view.Dir;
	        axis.normalize();
	    }  
            // Normalize
            v0.normalize();
	    v1.normalize();
	    // Third axis, normal to the other two
            n0 = v3d.cross(v0,axis);
	    n1 = v3d.cross(v1,axis);
	    // Now, do the rotation of the image vectors.
            view.Dir = v3d.dot(view.Dir,v0)*v1+
		       v3d.dot(view.Dir,n0)*n1+
		       v3d.dot(view.Dir,axis)*axis;
	    view.Dx = v3d.dot(view.Dx,v0)*v1+
		      v3d.dot(view.Dx,n0)*n1+
		      v3d.dot(view.Dx,axis)*axis;
	    view.Dy = v3d.dot(view.Dy,v0)*v1+
		      v3d.dot(view.Dy,n0)*n1+
		      v3d.dot(view.Dy,axis)*axis;
	    // Update the view
            tileCache.invalidateView();
            daImageView.QueueDraw();        
	}
    }
    
    
}

// vim: shiftwidth=4 softtabstop=4 expandtab
