/*
 * Vector3D.cs: Basic 3d-vector class
 * Copyright (c) 2004 Daniel Serpell
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

namespace Vector3D {

    public struct v3d {
        public double x, y, z;
        public static v3d operator * ( double s, v3d v ) {
            v3d n = v;
            n.x*=s; n.y*=s; n.z*=s;
            return n;
        }
        public static v3d operator + ( v3d v1, v3d v2 ) {
            v3d n = v1;
            n.x += v2.x; n.y += v2.y; n.z += v2.z;
            return n;
        }
        public static v3d operator - ( v3d v1, v3d v2 ) {
            v3d n = v1;
            n.x -= v2.x; n.y -= v2.y; n.z -= v2.z;
            return n;
        }
        public static bool operator == ( v3d v1, v3d v2 ) {
            return (v1.x == v2.x) && (v1.y == v2.y) && (v1.z == v2.z);
        }
        public static bool operator != ( v3d v1, v3d v2 ) {
            return (v1.x != v2.x) || (v1.y != v2.y) || (v1.z != v2.z);
        }
        public override bool Equals(System.Object obj) {
            return obj is v3d && this == (v3d)obj;
        }
        public override int GetHashCode() {
            int i,j,k;
            i = x.GetHashCode();
            j = y.GetHashCode();
            k = z.GetHashCode();
            // TODO: do a better hash!
            return i ^ (j<<11) ^ (j>>21) ^ (k<<21) ^ (k>>11);
        }
        public double norm() {
            return System.Math.Sqrt(x*x+y*y+z*z);
        }
        public void normalize() {
            double n = 1.0 / norm();
            x *= n; y *= n; z *= n;
        }
        public override string ToString() {
            string s;
            s = "(" + x.ToString() + "," + y.ToString() + "," + 
               z.ToString() + ")";
            return s;
        }
        public static v3d scale ( double s, v3d v )
        {
            return s * v;
        }
    
        public static double dot ( v3d v1, v3d v2 ) {
            return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
        }
    
        public static v3d cross ( v3d v1, v3d v2 ) {
            v3d n;
            n.x = v1.y * v2.z - v1.z * v2.y;
            n.y = v1.z * v2.x - v1.x * v2.z;
            n.z = v1.x * v2.y - v1.y * v2.x;
            return n;
        }
    
    };
    
}
