/*
 * ViewData.cs: Class to hold all view parameters.
 * Copyright (c) 2004 Daniel Serpell
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using Vector3D;

public class ViewData {

    protected v3d dir, dx, dy;
    protected double zoom;
    protected double xAxisCurvature;
    protected double yAxisCurvature;
    protected int tileSize;
    protected bool useDiag;
    protected bool updated;
    
    public int sx, sy;
    
    // Contructor
    public ViewData()
    {
        setAzimutElevRot(0,0,0);
      	zoom = 1;
	tileSize = 16;

    }
        
    // Property accesors
    public v3d Dir {
        get {
            return dir;
        }
        set {
            if( dir != value )
            {
                dir = value;
                notifyUpdate();
            }
        }
    }
    
    public v3d Dx {
        get {
            return dx;
        }
        set {
            if( dx != value )
            {
                dx = value;
                notifyUpdate();
            }
        }
    }
    
    public v3d Dy {
        get {
            return dy;
        }
        set {
            if( dy != value )
            {
                dy = value;
                notifyUpdate();
            }
        }
    }
    
    public double Zoom {
        get {
            return zoom;
        }
        set {
            if( zoom != value )
            {
                zoom = value;
                notifyUpdate();
            }
        }
    }
   
    public bool UseDiag {
        get {
            return useDiag;
        }
        set {
            if( useDiag != value )
            {
                useDiag = value;
                notifyUpdate();
            }
        }
    }
   
    public int TileSize {
        get {
            return tileSize;
        }
        set {
            if( tileSize != value )
            {
                tileSize = value;
                notifyUpdate();
            }
        }
    }
    
    public double XAxisCurvature {
        get {
            return xAxisCurvature;
        }
        set {
            if( xAxisCurvature != value )
            {
                xAxisCurvature = value;
                notifyUpdate();
            }
        }
    }
    
    public double YAxisCurvature {
        get {
            return yAxisCurvature;
        }
        set {
            if( yAxisCurvature != value )
            {
                yAxisCurvature = value;
                notifyUpdate();
            }
        }
    }
    
    // Events
    public event System.EventHandler NotifyUpdate;
    
    // Event notifier
    protected void notifyUpdate()
    {
        if (NotifyUpdate != null) NotifyUpdate(this, new System.EventArgs());
    }
    
    // Methods
    public void setAzimutElevRot(double az, double el, double rot)
    {
        double ca = System.Math.Cos(az);
        double sa = System.Math.Sin(az);
        double ce = System.Math.Cos(el);
        double se = System.Math.Sin(el);
        double cr = System.Math.Cos(rot);
        double sr = System.Math.Sin(rot);
        dir.x = sa * ce;
        dir.z = ca * ce;
        dir.y = se;
        dx.x = ca*cr+sa*se*sr;
        dx.y = -ce*sr;
        dx.z = -sa*cr+ca*se*sr;
        dy = v3d.cross(dir,dx);
        notifyUpdate();
    }
    
    public void getAzimutElevRot(out double az,out double el,out double rot)
    {
        double te,ta,tr;
        // Normalize
        dir.normalize();
        dx.normalize();
        double t = dir.x * dir.x + dir.z * dir.z;
        if( t > 0 )
        {
            // Elevation
            el = System.Math.Asin( dir.y );
            // Azimuth
            az = System.Math.Atan2( dir.x, dir.z );
            if( az < 0 )
                az += 2*System.Math.PI;
            // Rotation
            rot = System.Math.Asin( - dx.y / System.Math.Sqrt(t));
            if( rot < 0 )
                rot += 2*System.Math.PI;
        }
        else
        {
            // Elevation
            el = System.Math.Asin( dir.y );
            // Azimuth
            az = System.Math.Atan2( dx.z, dx.x );
            if( az < 0 )
                az += 2*System.Math.PI;
            // Rotation
            rot = 0;
        }
    }
    
    // Set zoom
    public void zoomIn() {
        zoom *= 1.25;
        notifyUpdate();
    }
    
    public void zoomOut() {
        zoom *= 0.8;
        notifyUpdate();
    }
    
    // Set different view-points
    public void goNorth() {
        setAzimutElevRot(0,0,0);
    }
    
    public void goNorthEast() {
        setAzimutElevRot(System.Math.PI*0.25,0,0);
    }
    
    public void goEast() {
        setAzimutElevRot(System.Math.PI*0.5,0,0);
    }
    
    public void goSouthEast() {
        setAzimutElevRot(System.Math.PI*0.75,0,0);
    }
    
    public void goSouth() {
        setAzimutElevRot(System.Math.PI,0,0);
    }
    
    public void goSouthWest() {
        setAzimutElevRot(System.Math.PI*1.25,0,0);
    }
        
    public void goWest() {
        setAzimutElevRot(System.Math.PI*1.5,0,0);
    }
    
    public void goNorthWest() {
        setAzimutElevRot(System.Math.PI*1.75,0,0);
    }
    
    public void goDown() {
        setAzimutElevRot(0,-System.Math.PI*0.5,0);
    }
    
    public void goUp() {
        setAzimutElevRot(0,System.Math.PI*0.5,0);
    }
    
}
